#-----------------------------------------
# EOO_Calculator_v1.5.py
# Version:1.5
# Original: Adrian Hughes IUCN
# Edited: Jemma Able IUCN 21/07/2015
# 26/04/16 - updated delete_management from arcpy.Delete_management(feature_eoo) to arcpy.Delete_management(feature_eoo, "")
# May 2021 Kate Harding IUCN - to implement change of 'binomial' to 'sci_name'
# v1.5: Checks that output workspace is a geodatabase and exits if it is not
#   Checks for coordinate system and exits with warning message if "Unknown" 
#   Enhanced check for P/O/S fields to list missing field(s) in message
#   Uses input species name field for the output field name (to accommodate changeover period from binomial to sci_name 
#
# Usage: This script will only work with ArcGIS 10.3 or above advanced version. 
#   The convex hull method of the ESRI minimum bounding geometry tool requires the advanced version
#
# The script calculates EOO by creating a convex hull minimum bounding geometry for each species range with the specified hardcoded mapping codes (see below)
# The data is then projected to the cylindrical equal area projected coordinate system and the area of the EOO is
#   calculated in km2 and added to the attribute data of the output for each species
# EOO is calculated around polygons or points with attribute values satisfying
#   presence 1 or 2; origin 1, 2 or 6; seasonal 1 & 2 or 1 & 3
# If user unchecks 'use IUCN attributes' checkbox then attribute values are ignored and EOO is calculated from all input data
# Warning: Tool can sometimes crash ArcMap if large complex geometries are used


import IUCNSP
import shutil
import sys
import os
import traceback
import inspect
import arcpy
from arcpy import env
import datetime

arcpy.env.overwriteOutput = True

try:          

    # Parameters    
    Input = arcpy.GetParameterAsText(0)
    speciesfield = arcpy.GetParameterAsText(1)
    workspace = arcpy.GetParameter(2)
    use_attributes = arcpy.GetParameterAsText(3)
    summary_table = arcpy.GetParameterAsText(4)

    # Variables
    S3count = 0

    # Setting workspace and get default spatial ref file on disk
    arcpy.env.workspace = workspace
    # output location must be geodatabase
    if str(workspace)[-4:] <> ".gdb":
        arcpy.AddWarning("Please specify a geodatabase for the output location.\n{} is not a geodatabase".format(workspace))
        sys.exit(0)

    # check that the input species data has a defined coordinate system
    sr = arcpy.Describe(Input).spatialreference.name
    if sr == "Unknown":
        msg = "No coordinate system for input dataset.\nPlease check! \nData should be in GCS_WGS_1984"
        print msg
        arcpy.AddError(msg)
        sys.exit(0)

    # get the Cylindrical Equal Area projection file from tooldata folder
    scriptpath = sys.argv[0]
    tooldir = os.path.split(os.path.split(scriptpath)[0])[0] # move 2 levels up path
    CEApath = os.path.join(tooldir, 'tooldata', r'Cylindrical Equal Area (world).prj')  
    arcpy.AddMessage("CEApath = {}".format(CEApath))
       
    # get a unique list of values/species names and sort    
    specieslist = IUCNSP.GetUniqueValuesFromShapefile(Input, speciesfield)
    specieslist.sort()

    # Check if specieslist is null
    if not specieslist:
        msg = "\nNo taxa found! Exiting the tool.\n"
        print msg
        arcpy.AddWarning(msg)
        sys.exit(2)
    
    # if using IUCN P/O/S codes, check form required fields
    if str(use_attributes) == 'true':
        IUCNSP.Printboth("Checking for IUCN P/O/S attributes...")
        missing = []
        if not IUCNSP.CheckFieldExists(Input,"PRESENCE"):
            missing.append("PRESENCE")
        if not IUCNSP.CheckFieldExists(Input,"ORIGIN"):
            missing.append("ORIGIN")
        if not IUCNSP.CheckFieldExists(Input,"SEASONAL"):
            missing.append("SEASONAL")
        if len(missing) > 0:
            msg = "IUCN field(s) {} missing from input data! \nPRESENCE, ORIGIN and SEASONAL must all be present if using IUCN attributes.".format(missing)
            print msg
            arcpy.AddWarning(msg)
            msg = "Add the missing attribute(s)\nor run without IUCN attributes.\nExiting the tool\n"
            print msg
            arcpy.AddWarning(msg)
            sys.exit(0)

    # process each species data    
    for species in specieslist:      
        IUCNSP.Printboth("\nProcessing {}".format(species))
        species_valid = IUCNSP.CleanSpeciesName(species)      
        feature_eoo =  species_valid + "_EOO"
        feature_eoo_proj =  species_valid + "_EOO_proj"
        feature_eoo_nb_proj = species_valid + "_EOO_NB_proj"
        
        
       # set default where clause
        if str(use_attributes) == 'true':
            defwhereclause =  "{} = '{}' AND \"PRESENCE\"  IN (1,2) AND \"ORIGIN\" IN (1,2,6) AND \"SEASONAL\" IN (1,2)".format(speciesfield,species)              
        else:
            defwhereclause = "{} = '{}'".format(speciesfield,species)  
            
        arcpy.MakeFeatureLayer_management(Input, "lyr", defwhereclause)       
                    

        # Calculate min bounding geometry
        IUCNSP.Printboth("...calculating min bounding geometry")
        arcpy.MinimumBoundingGeometry_management("lyr",feature_eoo,"CONVEX_HULL", "ALL")

        # Project to cylindrical equal area to calculate area
        IUCNSP.Printboth("...projecting and calculating area")
        arcpy.Project_management(feature_eoo,feature_eoo_proj, CEApath)


        # Add field to store scientific name and area in sqkm
        IUCNSP.AddFieldWithValue(feature_eoo_proj, speciesfield, "TEXT", "'" + species + "'")
        IUCNSP.AddField(feature_eoo_proj, "EOO_AREA_SQKM", "FLOAT")        
        arcpy.CalculateField_management(feature_eoo_proj, "EOO_AREA_SQKM",'!SHAPE_Area! / 1000000', "PYTHON")
        arcpy.Delete_management(feature_eoo, "")

        
        # If user checked use IUCN attributes box, calculate Non-breeding range
        # check whether species has Non-breeding data (Seasonal = 3)
        if str(use_attributes) == 'true':
            defwhereclause =  "{} = '{}' AND  \"SEASONAL\" = 3".format(speciesfield, species)  
            arcpy.MakeFeatureLayer_management(Input, "nblyr", defwhereclause)
            result = arcpy.GetCount_management("nblyr")
            S3count= int(result.getOutput(0)) 
            # if Non-breeding data exists, calculate EOO for non-breeding range  
            if S3count > 0:
                IUCNSP.Printboth("Non-breeding data exists for this species. Calculating Non Breeding EOO...")
                defwhereclause =  "{} = '{}' AND \"PRESENCE\"  IN (1,2) AND \"ORIGIN\" IN (1,2,6) AND \"SEASONAL\" IN (1,3)".format(speciesfield,species)    
                arcpy.MakeFeatureLayer_management(Input, "nblyr", defwhereclause)

                # Calculate min bounding geometry
                IUCNSP.Printboth("...calculating min bounding geometry")
                arcpy.MinimumBoundingGeometry_management("nblyr",feature_eoo,"CONVEX_HULL", "ALL")

                # Project to cylindrical equal area to calculate area
                IUCNSP.Printboth("...projecting and calculating area for non-breeding distribution")
                arcpy.Project_management(feature_eoo,feature_eoo_nb_proj, CEApath)

                # Add field to store scientific name and area in sqkm
                IUCNSP.AddFieldWithValue(feature_eoo_nb_proj, speciesfield, "TEXT", "'{}'".format(species))
                IUCNSP.AddField(feature_eoo_nb_proj, "EOO_NB_SQKM", "FLOAT")        
                arcpy.CalculateField_management(feature_eoo_nb_proj, "EOO_NB_SQKM",'!SHAPE_Area! / 1000000', "PYTHON")
                arcpy.Delete_management(feature_eoo, "")  


    #create summary table of all species
    if str(summary_table) == 'true':
        IUCNSP.Printboth("\nCreating Summary Table")
        fcList = arcpy.ListFeatureClasses("*_EOO_*","polygon")
        arcpy.Merge_management(fcList, "SpeciesMerge")

        statsFieldsNB = [["EOO_AREA_SQKM", "MIN"], ["EOO_NB_SQKM", "MIN"]]
        statsFields = [["EOO_AREA_SQKM", "MIN"]]
    
        if IUCNSP.CheckFieldExists("SpeciesMerge","EOO_NB_SQKM"):
            IUCNSP.Printboth("There are taxa with non-breeding areas")
            arcpy.Statistics_analysis("SpeciesMerge","SpeciesEOOSummary", statsFieldsNB, speciesfield)
        else:
            IUCNSP.Printboth("No taxa with non-breeding areas")
            arcpy.Statistics_analysis("SpeciesMerge","SpeciesEOOSummary", statsFields, speciesfield)
        arcpy.Delete_management("SpeciesMerge", "")
        arcpy.DeleteField_management("SpeciesEOOSummary", "FREQUENCY")

except SystemExit as err:
    pass

except arcpy.ExecuteError: 
    # Get the tool error messages 
    # 
    msgs = arcpy.GetMessages(2) 

    # Return tool error messages for use with a script tool 
    #
    arcpy.AddError(msgs) 

    # Print tool error messages for use in Python/PythonWin 
    # 
    print msgs

except:
    # Get the traceback object
    #
    tb = sys.exc_info()[2]
    tbinfo = traceback.format_tb(tb)[0]

    # Concatenate information together concerning the error into a message string
    #
    pymsg = "PYTHON ERRORS:\nTraceback info:\n" + tbinfo + "\nError Info:\n" + str(sys.exc_info()[1])
    msgs = "ArcPy ERRORS:\n" + arcpy.GetMessages(2) + "\n"

    # Return python error messages for use in script tool or Python Window
    #
    arcpy.AddError(pymsg)
    arcpy.AddError(msgs)

    # Print Python error messages for use in Python / Python Window
    #
    print pymsg + "\n"
    print msgs
   
